//
//  ABOFloatVector.m
//  Magi
//
//  Created by Jamie Hardt on 4/8/12.
//  Copyright (c) 2012 Jamie Hardt. All rights reserved.
//

#import "ABOSampleVector.h"

uint_t get_frame_count(AudioBufferList bufferList, 
                       AudioStreamBasicDescription streamDesc) {
    uint_t retVal = 0;
    
    if (bufferList.mNumberBuffers > 0 ) {
        if((streamDesc.mFormatFlags & kAudioFormatFlagIsNonInterleaved)) {
            retVal = bufferList.mBuffers[0].mDataByteSize / streamDesc.mBytesPerFrame;
            // I'm not sure this works, not sure if each sample is a frame under this scenario
        } else {
            NSUInteger i;
            for (i = 0; i < bufferList.mNumberBuffers; i++) {
                retVal += bufferList.mBuffers[i].mDataByteSize / streamDesc.mBytesPerFrame;
            }
        }
    }
    
    return retVal;
}

#define ASSERT_GOOD_BOUNDS(len,ch) do {\
NSAssert(len < self.length && ch < self.channelCount,\
@"%s, Frame/channel out of bounds",_cmd);\
} while (0);

@implementation ABOSampleVector

@synthesize FVec = _priv;

#pragma mark Initializers

-(id)initWithCMSampleBuffer:(CMSampleBufferRef)sampleBuffer {
    AudioBufferList bufList;
    const AudioStreamBasicDescription *streamDesc;
    CMBlockBufferRef blockBuffer = NULL;
    
    NSAssert(CMSampleBufferDataIsReady(sampleBuffer), 
             @"bad sample buffer, data is not ready");
    
    CMFormatDescriptionRef formatDesc = CMSampleBufferGetFormatDescription(sampleBuffer);
    
    NSAssert(CMFormatDescriptionGetMediaType(formatDesc) == kCMMediaType_Audio,
             @"sampleBuffer does not contain audio-only data");
    
    NSAssert(
    CMFormatDescriptionGetMediaSubType(formatDesc) == (FourCharCode)'lpcm'
    ,
             @"sampleBuffer does not contain lpcm data");
    
    streamDesc = CMAudioFormatDescriptionGetStreamBasicDescription(formatDesc);
    
    CMSampleBufferGetAudioBufferListWithRetainedBlockBuffer(sampleBuffer, 
                                                            NULL, 
                                                            &bufList, 
                                                            sizeof(bufList), 
                                                            kCFAllocatorDefault, 
                                                            kCFAllocatorDefault, 
                                                                kCMSampleBufferFlag_AudioBufferList_Assure16ByteAlignment, 
                                                            &blockBuffer);
    
    //CFRetain(blockBuffer);
    
    self = [self initWithAudioBufferList:bufList
                  audioStreamBasicDescription:*streamDesc];
    
    CFRelease(blockBuffer);
    
    return self;
}

-(id)initWithAudioBufferList:(AudioBufferList)bufferList 
 audioStreamBasicDescription:(AudioStreamBasicDescription)streamDesc {
    NSAssert(streamDesc.mFormatID == kAudioFormatLinearPCM, 
             @"bad format, needs to be kAudioFormatLinearPCM");
    
    NSAssert(streamDesc.mFormatFlags & kAudioFormatFlagIsFloat, 
             @"bad format, needs to be kAudioFormatFlagIsFloat");
    
    NSAssert(!(streamDesc.mFormatFlags & kAudioFormatFlagIsNonInterleaved), 
             @"bad format, needs to be !kAudioFormatFlagIsNonInterleaved");
    
    NSAssert(streamDesc.mFramesPerPacket == 1,
             @"bad format, frames per packet must == 1");
    
    NSAssert(streamDesc.mBytesPerFrame == streamDesc.mChannelsPerFrame * sizeof(float),
             @"bad format, frames must be made of floats");
    
    self = [self initWithLength:get_frame_count(bufferList,streamDesc)
                   channelCount:streamDesc.mChannelsPerFrame];
    if (self) {
        NSUInteger i, j, k, frame = 0;
        for (i = 0; i < bufferList.mNumberBuffers; i++) {
            AudioBuffer thisBuffer = bufferList.mBuffers[i];
            
            j = 0;
            while (j < thisBuffer.mDataByteSize) {
                
                for (k = 0; k < thisBuffer.mNumberChannels ; k++) {
                    float thisSample = 0.0f;
                    memcpy(&thisSample, thisBuffer.mData + j + k * sizeof(float), sizeof(float));
                    [self setSample:thisSample atFrame:frame inChannel:k];
                }
                frame++;
                
                j += streamDesc.mBytesPerFrame;
            }
        }
        
    }

    return self;
}


-(id)initWithLength:(uint_t)length 
       channelCount:(uint_t)channels {
    
    fvec_t *v = new_fvec(length, channels);
    if (v) {
        self = [self initWithFvecNoCopy:v freeWhenDone:YES];
    } else {
        self = nil;
    }
    
    return self;
}

-(id)initWithFvec:(fvec_t *)vector {
    NSAssert(vector != NULL, @"tried to init with null vector ");
    self = [super init];
    if (self) {
        _priv = new_fvec(vector->length, vector->channels);
        if (!_priv) {
            return nil;
        }
        int chan;
        for (chan = 0; chan < vector->channels; chan++) {
            int i;
            for (i = 0; i < self.length; i++) {
                _priv->data[chan][i] = vector->data[chan][i];
            }
        }
        
        _privFreeWhenDone = YES;
    }
    return self;
}

-(id)initWithFvecNoCopy:(fvec_t *)vector  freeWhenDone:(BOOL)freeWhenDone {
    NSAssert(vector != NULL,@"tried to init with a NULL fvec_t");
    self = [super init];
    if (self) {
        _priv = vector;
        _privFreeWhenDone = freeWhenDone;
    }
    return self;
}

- (void)dealloc {
    if (_priv && _privFreeWhenDone) { del_fvec(_priv); };
}

#pragma mark Copying

-(id)copyWithZone:(NSZone *)zone {
    return [[ABOSampleVector allocWithZone:zone] initWithFvec:_priv];
}

#pragma mark Coding

-(void)encodeWithCoder:(NSCoder *)aCoder {
    
    [aCoder encodeInteger:self.channelCount forKey:@"channelCount"];
    [aCoder encodeInteger:self.length forKey:@"length"];
    NSMutableArray *chanData = [NSMutableArray array];
    
    int i;
    for (i = 0; i < self.channelCount; i++) {
        [chanData addObject:[NSData dataWithBytes:_priv->data[i]
                                           length:(self.length * sizeof(smpl_t))]];
    }
    [aCoder encodeObject:chanData forKey:@"channelData"];
}

-(id)initWithCoder:(NSCoder *)aDecoder {
    self  = [self initWithLength:[aDecoder decodeIntegerForKey:@"length"] 
                    channelCount:[aDecoder decodeIntegerForKey:@"channelCount"]];
    if (self) {
        NSArray *channelData = [aDecoder decodeObjectForKey:@"channelData"];
        
        int i = 0;
        for (NSData *aChannel in channelData) {
            [aChannel getBytes:_priv->data[i++] length:self.length * sizeof(smpl_t)];
        }
    }
    
    return self;

}

#pragma mark -
#pragma mark Composition


-(id)vectorByAppendingVector:(ABOSampleVector *)vector {
    NSAssert( vector.channelCount == self.channelCount,@"channel counts do not match");
    
    ABOSampleVector *retObj = [[ABOSampleVector alloc] 
                               initWithLength:(self.length + vector.length)
                               channelCount:self.channelCount];
    
    int chan;
    for (chan = 0 ; chan < retObj.channelCount; chan++) {
        [retObj putSamples:[self getChannel:chan]
                withLength:self.length
                  atOffset:0 
                 inChannel:chan];
        
        [retObj putSamples:[vector getChannel:chan]
                withLength:vector.length
                  atOffset:self.length
                 inChannel:chan];
    }

    return retObj;
}

-(id)vectorByTrimmingFromSample:(uint_t)offset length:(uint_t)length {
    NSAssert(offset + length <= self.length,@"trim extent %u exceeds bounds of vector %u",
             offset+length,self.length);
    
    ABOSampleVector *retObj = [[ABOSampleVector alloc] initWithLength:length
                                                         channelCount:self.channelCount];
    int chan;
    for (chan = 0 ; chan < retObj.channelCount; chan++) {
        [retObj putSamples:&[self getChannel:chan][offset]
                withLength:length
                  atOffset:0
                 inChannel:chan];
    }
    
    return retObj;
}

-(NSArray *)subvectorsOfLength:(NSUInteger)length
                        period:(NSUInteger)period {
    NSUInteger i, window;
    NSMutableArray *retObj = [NSMutableArray array];
    for (i = 0; i < self.length; i += period) {
        window = MIN( self.length - i , length );
        [retObj addObject:[self vectorByTrimmingFromSample:i length:window]];
    }
    
    return [NSArray arrayWithArray:retObj];
}


#pragma mark Properties

-(uint_t)length {
    return _priv->length;
}

-(uint_t)channelCount {
    return _priv->channels;
}

#pragma mark Work with Samples

-(smpl_t)sampleAtFrame:(uint_t)frame inChannel:(uint_t)chan {
    ASSERT_GOOD_BOUNDS(frame, chan);

    return _priv->data[chan][frame];
}

-(void)setSample:(smpl_t)sampleValue 
         atFrame:(uint_t)frame inChannel:(uint_t)chan {
    ASSERT_GOOD_BOUNDS(frame, chan);

    
    _priv->data[chan][frame] = sampleValue;
    
}

#pragma mark Work with Buffers of Samples

-(smpl_t *)getChannel:(uint_t)channel {
    return _priv->data[channel];
}

-(void)putSamples:(smpl_t *)samples withLength:(uint_t)length
         atOffset:(uint_t)offset 
        inChannel:(uint_t)channel {
    ASSERT_GOOD_BOUNDS(offset, channel);
    NSAssert(offset + length <= self.length,@"put sample length %u exceeds bounds of vector %u",length,self.length);

    memcpy( &_priv->data[channel][offset] , 
           samples, 
            length * sizeof(smpl_t));
}

-(void)putSamples:(smpl_t *)samples withLength:(uint_t)length
        inChannel:(uint_t)channel {
    
    [self putSamples:samples withLength:length 
            atOffset:0 inChannel:channel];
}


@end
